////////////////////////////////////////////////////////////////////////
//
// 46-927 Data Structures, Algorithms and AI
//
// File: AbstractOption.cpp
//
// Student Name:            
// Date:         
// Comments:      Implementation of an abstract Option class
//              
//
//                This class is abstract.  Therefore, you must
//                derive a class from this class and add code
//                for the pure virtual functions declared in
//                AbstractOption.h
//
//                Chapter 2 and Chapter 3 of the course notes
//                describe pricing an option on a PDAG.  This an 
//                implementation of the process they describe.
//
//
//                ***
//                For this assignment you must write Calc_American_Option
//                and Calc_European_Rand().  See Sections 3.2 and 3.3
//                of the course notes.
//                ***
//
//
//                See AbstractOption.h for more detail
//
////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <time.h>
#include <iostream.h>
#include "AbstractOption.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAbstractOption::CAbstractOption(int T, double r, double u, double d, double s_0)
{

    m_timeHorizon = T;
    m_rate = r;
    m_upFactor = u;
    m_downFactor = d;
    m_initial = s_0;

    //calculate the risk-nuetral probabilities
    m_probUp = (1+ m_rate - m_downFactor)/(m_upFactor - m_downFactor);
    m_probDown = 1-m_probUp;

    //Allocate the link list
    m_nodes = new CLinkedList[m_timeHorizon+1];

    srand((unsigned)time(NULL)); 
   
}

CAbstractOption::~CAbstractOption()
{
    delete [] m_nodes;
}


void CAbstractOption::Calc_European_Option()
{
    int time;
    double sum;
    CKey aKey;
    CNode* aNode;

    if (TESTDEBUG) 
    {
        cout << "Entered Calc_European_Option" << endl;
    }

    //Start at time = m_timeHorizon, work backwards
    for(time=m_timeHorizon; time >= 0; time--) 
    {
        aNode = m_nodes[time].m_head;
        while (aNode != NULL) 
        {
	        if (time == m_timeHorizon) 
            {
                aKey = aNode->GetData();
                aKey.SetOptionValue(Payoff(aKey.GetStockPrice(), aKey.GetAuxProcess(), m_timeHorizon));
                aNode->SetData(aKey);
            }
	        else 
            {
	 
                aKey = aNode->GetData();

                sum = m_probUp * aNode->GetSuccessor(0)->GetData().GetOptionValue() +
                      m_probDown* aNode->GetSuccessor(1)->GetData().GetOptionValue();

                aKey.SetOptionValue(sum/(1+m_rate));

                aNode->SetData(aKey);

	        }
	        aNode=aNode->GetNext();
        }
    }

    if (TESTDEBUG)
    {
        cout << "Leaving Calc_European_Option" << endl;
    }
}

void CAbstractOption::GenPDAG()
{
    CNode* aNode;
    double s;
    double a;
    int time;

    if (TESTDEBUG) 
    {
        cout << "Entered GenPDAG" << endl;
    }

    m_nodes[0].InsertUnique(CKey(0,m_initial,m_initial));

    for(time=0; time < m_timeHorizon; time++) 
    {
        aNode=m_nodes[time].m_head;

        while (aNode != NULL) 
        {
	        if (TESTDEBUG) 
            {
	            cout<< "Considering node: " << endl;
                aNode->PrintNode();
	        }
	  
	        s = aNode->GetData().GetStockPrice();
	        a = aNode->GetData().GetAuxProcess();
            
	        //Generate successor nodes
            CKey up_key(time+1, s * m_upFactor, F(s * m_upFactor, a));
            CKey down_key(time+1, s * m_downFactor, F(s * m_downFactor, a));


	        aNode->SetSuccessor(0, m_nodes[time+1].InsertUnique(up_key));
	        aNode->SetSuccessor(1, m_nodes[time+1].InsertUnique(down_key));
	
            aNode = aNode->GetNext();
	      
        }

        if (TESTDEBUG)
        {
            PrintPDAG();
        }

    }

	
    if (TESTDEBUG) 
    {
        cout <<"Leaving GenPdag" << endl;
    }

}


void CAbstractOption::PrintPDAG()
{
    int time;

    for(time=0; time <= m_timeHorizon; time++) 
    {
        cout << "Nodes at time " << time << endl;
        cout << "------------BEGIN-----------------" << endl;
        m_nodes[time].PrintList();
        cout << "------------END-------------------" << endl;
    }

}

void CAbstractOption::MonteCarloPath(CNode** pathArray)
{

    //For each node, m_probUp is probability of taking the 
    //edge to successor 0, m_probDown is the probability of
    //taking the edge to successor 1

    int i;
    CNode* m_current;
    double randVal; 

    m_current = m_nodes[0].m_head;

    for (i=0; i<(m_timeHorizon); i++)
    {
        pathArray[i]=m_current;
        
        randVal = ((rand() * 1.0) / RAND_MAX);

        if (randVal <= m_probUp)
        {
            m_current = m_current->GetSuccessor(0);
        }
        else
        {
            m_current = m_current->GetSuccessor(1);
        }
    }

    pathArray[m_timeHorizon] = m_current;

}

void CAbstractOption::PrintPaths(int numPaths)
{
    int i;
    int j;
    CNode** path;
    path = new CNode*[m_timeHorizon+1];

    for (i=0; i<numPaths; i++)
    {
        MonteCarloPath(path);
        cout << endl;
        cout << "Path " << i <<endl;
        cout << "------------------------------" << endl;
        for (j=0; j<=m_timeHorizon; j++)
        {
            path[j]->PrintNode();
        }
        cout << "End Of Path " << i << endl;
    }

    //not deleting nodes, just deleting an array of pointers
    delete [] path;
}


int CAbstractOption::CountLatticeNodes()
{
    int time;
    int numNodes = 0;

    for(time=0; time <= m_timeHorizon; time++) 
    {
        numNodes = numNodes + m_nodes[time].CountNodes();
    }

    return numNodes;

}

double CAbstractOption::GetFinalOptionValue()
{
    return m_nodes[0].m_head->GetData().GetOptionValue();
}



